//swfunc.c

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h> 
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>    
#include <time.h>
#include <pthread.h>
#include <signal.h>
#include "myws.h"
#include "wsserver.h"
#include "wsfunc.h"
#include "config.h"
#include <openssl/sha.h>
#include "base64.h"


/* === выборка значений ключей из строк ===
inbuf  - исходная строка для разбора
inref  - имя ключа в исходной строке
output - вывод значения ключа
!!! - имя ключа и его значение в исходной строке засеняются 
пробелами
*/
__inline int strval(char * inbuf, char * inref, char * output)
{
	char		* p;
	uint8_t		m = 0;
	uint8_t		n = 0;
	uint8_t		kflag = 0;
	if((p = strstr(inbuf, inref)) != NULL)
	{
		while(1)
		{
			if(((*(char*)(p + m)) == '\n') ||
			((*(char*)(p + m)) == '\r'))
				break;
			else if((*(char*)(p + m)) == ' ')
				kflag = 0xFF;
			else if(kflag)
			{
				(*(char*)(output + n)) = (*(char*)(p + m));
				n++;
			}
			(*(char*)(inbuf + m)) = ' ';
			m++;
		}
		return 1;
	}
	else
		return 0;
}

/* === выделение параметров запроса === */
int wsin(WSDAT * wsd)
{
	int res;
	char	outstr[64];
	printf("Rd:\n%s\n", (char*)(wsd->buffer));
	printf("=== wsin ===\n");
	memset(wsd->Sec_WebSocket_Key, 0, sizeof(wsd->Sec_WebSocket_Key));
	res = strval((char*)(wsd->buffer), "Sec-WebSocket-Key:", (char*)(wsd->Sec_WebSocket_Key));
	if(res)
		printf("Sec-WebSocket-Key: %s\n", wsd->Sec_WebSocket_Key);
	memset(&outstr[0], 0, 64);
	res = strval((char*)(wsd->buffer), "Sec-WebSocket-Version:", (char*)(&outstr[0]));
	wsd->Sec_WebSocket_Version	= atoi(&outstr[0]);
	if(res)
		printf("Sec-WebSocket-Version: %i\n", wsd->Sec_WebSocket_Version);
	memset(wsd->Sec_WebSocket_Extensions, 0, sizeof(wsd->Sec_WebSocket_Extensions));
	res = strval((char*)(wsd->buffer), "Sec-WebSocket-Extensions:", (char*)(wsd->Sec_WebSocket_Extensions));
	if(res)
		printf("Sec-WebSocket-Extensions: %s\n", wsd->Sec_WebSocket_Extensions);
	//
	return res;
}

/* === формирование данных для отправки === */
int wsout(WSDAT * wsd)
{
	char		outstr[64];
	char		temp[_SHA_DIGEST_LENGTH] = {0};
	SHA1_CTX	context;
	memset(&outstr, 0, sizeof(outstr));
	printf("=== wsout ===\n");
	strcpy(&outstr[0], wsd->Sec_WebSocket_Key);
	strcat(outstr, GUIDKey);
	printf("::: %s\n", &outstr[0]);
	//SHA1(&temp[0], &outstr[0], strlen(outstr)-1 );
	SHA1((char*)(&outstr[0]), strlen(outstr), temp);
	printf("SHA1: %s\n", &temp[0]);
	memset(wsd->Sec_WebSocket_Accept, 0, sizeof(wsd->Sec_WebSocket_Accept));
	Base64encode(wsd->Sec_WebSocket_Accept, temp, strlen(temp));
	printf("Key_for_client: %s\n", wsd->Sec_WebSocket_Accept);
	memset(wsd->resp, 0, sizeof(wsd->resp));
	snprintf(wsd->resp, 130, "%s%s%s", response_ws, wsd->Sec_WebSocket_Accept, "\r\n\r\n");
	wsd->szresp = strlen(wsd->resp);
	printf("Resp(size=%i):\n%s\n", wsd->szresp, wsd->resp);
	//
	return 0;
}

/*
секция работы с websocket
*/

int wsfunc(WSDAT * wsd)
{
	int		res;
	uint8_t		opcode;
	uint8_t		payload_len;
	uint8_t		mask;
	uint8_t		rdbuf[_BUFSIZE] = {0};
	printf("=== wsfunc ===\n");
	if(wsd->flagping)
	{
		wsd->flagping	= 0;
		memset(wsd->buffer, 0, _BUFSIZE);
		wsd->buffer[0]	= 0x89;
		wsd->buffer[1]	= 0x05;
		strcpy((wsd->buffer + 2), "Hello");
		return 7;
	}
	else
	{
		printf("===\nRd(ws):\n%s\n==\n", wsd->buffer);
		opcode 		= wsd->buffer[0] & 0x0F;
		payload_len	= wsd->buffer[1] & 0x7F;
		mask		= wsd->buffer[1] & 0x80 ? 0xFF:0x00;
		//
		printf("FIN: 0x%02X\n", opcode & 0x01);
		printf("RSV1: 0x%02X\n", opcode & 0x02);
		printf("RSV2: 0x%02X\n", opcode & 0x04);
		printf("RSV3: 0x%02X\n", opcode & 0x08);
		printf("Opcode: 0x%02X\n", opcode & 0x0F);
		printf("payload_len: %i\n", payload_len);
		printf("Mask On: 0x%02X\n", mask);
		//
		wsd->masking_key[0] = wsd->buffer[2];
		wsd->masking_key[1] = wsd->buffer[3];
		wsd->masking_key[2] = wsd->buffer[4];
		wsd->masking_key[3] = wsd->buffer[5];
		//
		switch(opcode)
		{
			case WS_OPCODE_CONTINUE:
				printf("opcode = WS_OPCODE_CONTINUE\n");
				break;
			case WS_OPCODE_TEXT:
				printf("opcode = WS_OPCODE_TEXT\n");
				unsigned int i = 6, pl = 0;
        	    		for(; pl < payload_len; i++, pl++)
        	     		{
        	       			rdbuf[pl] = wsd->buffer[i]^wsd->masking_key[pl % 4]; 
        	     		}
				printf("\nReciv:%s\n", rdbuf);
				//
				res = strlen(wsd->buffer);
				return res;
			case WS_OPCODE_BINARY:
				printf("opcode = WS_OPCODE_BINARY\n");
				break;
			case WS_OPCODE_CLOSE:
				printf("opcode = WS_OPCODE_CLOSE\n");
				break;
			case WS_OPCODE_PING:
				printf("opcode = WS_OPCODE_PING\n");
				break;
			case WS_OPCODE_PONG:
				printf("opcode = WS_OPCODE_PONG\n");
				break;
			default:
				printf("opcode = ?\n");
				break;
		}
	}
	return 0;
}
