//wsserver.c

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h> 
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>    
#include <time.h>
#include <pthread.h>
#include <signal.h>
#include "wsserver.h"
#include "wsfile.h"
#include "wsfunc.h"
#include "config.h"


struct sockaddr_in		svr_addr, cli_addr;
socklen_t sin_len 	= 	sizeof(cli_addr);
int				wsport;
int				listener;
pthread_mutex_t         	serverWSmutex;
pthread_t               	hlistenWSprocess;
uint8_t				work;
WSSERVER			wsserver;

int	startWSserver(void)
{
	int one = 1;
	listener = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (listener < 0) 
	{
		printf("not socket.\n");
		return -1;
	}
	//
	setsockopt(listener, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(int));
	svr_addr.sin_family = AF_INET;
	svr_addr.sin_addr.s_addr = INADDR_ANY;
	svr_addr.sin_port = htons(wsport);
	//
	if(bind(listener, (struct sockaddr *) &svr_addr, sizeof(svr_addr)) == -1) 
   	{
		close(listener);
		printf("error bind.\n");
		return -2;
   	}
	if(listen(listener, _maxWSsocket) == -1) 
	{
		close(listener);
		printf("error listen.\n");
		return -3;
	}
	//mutex create
        if(!(pthread_mutex_init(&serverWSmutex,NULL) == 0))
                return 4;
        //starting listen
	work	= 0xFF;
        pthread_create(&hlistenWSprocess,NULL,(void*) serverWSlisten,NULL);
        sleep(1);       //wait one secoond
	//
	return 0;
}

void	serverWSlisten(void)
{
	uint8_t		w;
	int		err;
	int		count;
	int 		csock;
	fd_set 		fdss;
        struct timeval 	tvs;
	w	= work;
	memset(&wsserver.midcl[0], 0, _maxWSsocket);
	wsserver.cntclient	= 0;
	while(w)
	{
		tvs.tv_sec	= 0;
                tvs.tv_usec	= _SOCKET_TIMEOUT_;
                FD_ZERO(&fdss);
                FD_SET(listener,&fdss);
		//
		err = select(listener + 1,&fdss,&fdss,NULL,&tvs);
                if(err <= 0)
                {
			pthread_mutex_lock(&serverWSmutex);
			w	= work;
			pthread_mutex_unlock(&serverWSmutex);
                        continue;
                }
		csock = accept(listener,NULL,NULL);
		if(!(csock > 0))
                {
			pthread_mutex_lock(&serverWSmutex);
			w	= work;
			pthread_mutex_unlock(&serverWSmutex);
                        continue;
                }
		count	= 0;
		pthread_mutex_lock(&serverWSmutex);
                while(count < _maxWSsocket)
                {
                        if(wsserver.midcl[count] == 0)
                        {
				#if(_deb_con > 0)
					printf("ServCoon clien %i connecting...\n",count);
				#endif
                                wsserver.csock	 		= csock;
                                wsserver.idclient 		= count;
				wsserver.midcl[count]		= 0xFF;
				wsserver.cntclient++;
				//				
				pthread_attr_init(&wsserver.tattr[count]);
				pthread_attr_setdetachstate(&wsserver.tattr[count], PTHREAD_CREATE_DETACHED);
				pthread_create(&wsserver.pth[count], &wsserver.tattr[count],(void*) server_con, &wsserver);
				//
				break;
                        }
                        count++;
                }
                //
		w	= work;
                pthread_mutex_unlock(&serverWSmutex);
                usleep(_listen_sleep);
	}
	printf("!!! - End listen\n");
	pthread_exit(0); 
}

//чтение из сокета
__inline int wssRd(int nSocket, uint8_t *pBuffer, int size)
{
       //int     res                     = 1;
        int     	iRC                     = 0;
        int     	iReceiveStatus          = 0;
        int     	iStillToReceive         = size;
        int     	iCount 	= 0;
	//u_int16_t	sz	= 0;
        struct timeval 	ReceiveTimeout;
        fd_set 		fds;
	//u_int8_t	flag			= 0;
	//
        while(iStillToReceive > 0)
        {
                //установка величины таймаута
                FD_ZERO(&fds);
                FD_SET(nSocket, &fds);
                ReceiveTimeout.tv_sec  = 0;
                ReceiveTimeout.tv_usec = _SOCKET_TIMEOUT_;
                iRC = select(nSocket + 1, &fds, NULL, NULL, &ReceiveTimeout);
                //выход по таймауту
                if(!iRC)
		{
			if(iCount > 0)
				return iCount;
			else
				return -1;
		}
		//произошла какая то ошибка
                if(iRC < 0)
                        return -2;
                //прием нескольких байт
		iReceiveStatus = recv(nSocket, (u_int8_t*)(pBuffer), iStillToReceive, 0);
                if(iReceiveStatus >  0)
                {
			pBuffer 	+= iReceiveStatus;
                        iStillToReceive -= iReceiveStatus;
                        iCount 		+= iReceiveStatus;
                }
		else if(iReceiveStatus == 0)
		{
			if(iCount > 0)
				return iCount;
			else
				return -1;
		}
		else
			return -2;
        }
         return iCount;
}

//запись в сокет
__inline int wssWr(int nSocket, u_int8_t *pBuffer, u_int16_t iMessageLength)
{
 	int 		iCount		= 0;
        int 		iSC 		= 0;
        int 		iSendStatus 	= 0;
        struct timeval 	SendTimeout;
        fd_set 		fds;
	//fcntl(nSocket, F_SETFL, O_NONBLOCK);
        while(iMessageLength > 0)
        {
		//установка величины таймаута
        	SendTimeout.tv_sec  = 0;
        	SendTimeout.tv_usec = _SOCKET_TIMEOUT_;
        	FD_ZERO(&fds);
        	FD_SET(nSocket, &fds);
                iSC = select(nSocket + 1, NULL, &fds, NULL, &SendTimeout);
                if(!iSC)
                        return -1; //таймаут
                if(iSC < 0)
                        return -2; //ошибка
		//отправить несколько байт
		iSendStatus = send(nSocket, (char*)(pBuffer), iMessageLength, 0);
                if(iSendStatus > 0)
                {
                        //обновить буфер и счетчик
                        iMessageLength -= iSendStatus;
                        pBuffer += iSendStatus;
                        iCount += iSendStatus;
                }
		else if(iSendStatus == 0)
			return -1;
		else
			return -2;
        }
	return iCount;
}


//поток клиента
void server_con(WSSERVER * wss)
{
	WSSERVER	ws;
	WSDAT		wsdat;
	uint8_t		w;
	uint8_t		wsflag		= 0;
	int		client_fd;
	int		res;
	int		timelive	= _swtimelive;	
	pthread_mutex_lock(&serverWSmutex);
	w	= work;
	memcpy(&ws, wss, sizeof(WSSERVER));
	client_fd	= ws.csock;
	pthread_mutex_unlock(&serverWSmutex);
	printf("---client---\n");
	while(w)
	{
		memset(&wsdat.buffer[0], 0, _BUFSIZE);
		res = wssRd(client_fd, &wsdat.buffer[0], _BUFSIZE);
		switch(res)
		{
			case -1: //таймаут по приёму - делаем PING
				wsdat.flagping	= 0xFF;
				res = wsfunc(&wsdat);
				printf("Wr(ping) size=%i:%s\n", res, &wsdat.buffer[0]);
				res = wssWr(client_fd, &wsdat.buffer[0], res);
				break;
			case -2:
				goto mexit;
			default:
				if(!(wsflag))
				{
					res = wsin(&wsdat);
					res = wsout(&wsdat);
					res = wssWr(client_fd, &wsdat.resp[0], wsdat.szresp);
					if(res < 0)
						goto mexit;
					wsflag	= 0xFF;
				}
				else
				{
					wsdat.flagping	= 0;
					res = wsfunc(&wsdat);
					res = wssWr(client_fd, &wsdat.buffer[0], res);
				}
				timelive = _swtimelive;
				break;
		}
		pthread_mutex_lock(&serverWSmutex);
		w	= work;
		pthread_mutex_unlock(&serverWSmutex);
		if(timelive > 0)
			timelive -= _swclient_sleep;
		else
			goto mexit;
		usleep(_swclient_sleep);
	}
mexit:	shutdown(client_fd, SHUT_RDWR);
	pthread_mutex_lock(&serverWSmutex);
	wss->midcl[ws.idclient] = 0;
	wss->cntclient--;
	pthread_mutex_unlock(&serverWSmutex);
	printf("== Close Client ==\n");
	pthread_exit(0); 
}
